#include <stdio.h>

#include "sage++user.h"

#include "anls_att.h"

extern char * mixstring;

extern int getreflevel(SgSymbol * index, SgExpression * expr);
extern SgExpression * getlinerexp(int i, SgSymbol * symb, SgExpression * expr);

extern char * stackNameOfType(SgType *);
extern char * nameOfType(SgType *);
extern char * sizeNameOfType(SgType *);


Impdo_header_def::Impdo_header_def()
{
  number=0;
  for (int i=0; i<MAX_INDEX_DIM; i++) 
    for (int j=0; j<3; j++)
      index[i].indextriple[j]=NULL;

};

int Impdo_header_def::numberOfIndex()
{
  return number;
};

void Impdo_header_def::setIndexSymbol(SgSymbol * s)
{
  number++;
  index[number-1].indexvar=s;
};

SgSymbol * Impdo_header_def::indexSymbol(int i)
{
  return index[i].indexvar;
};

void Impdo_header_def::setIndexTriplet(int i, SgExpression * triplet)
{
  index[number-1].indextriple[i]=triplet;
};

SgExpression * Impdo_header_def::getIndexTriplet(int i, int j)
{
  return index[i].indextriple[j];
};

//------------------------;

io_reference::io_reference ()
{
  for (int i=0; i<MAX_ALIGN_DIM; i++) 
    for (int j=0; j<MAX_INDEX_DIM; j++)
      index_level[i][j]=0;




};

void io_reference::setExpression(SgExpression * e)
{
  expr=e;
};

SgExpression * io_reference::expression()
{
  return expr;
};

void io_reference::setIndex(int arraydim, int indexdim,int IL)
{
  index_level[arraydim][indexdim]=IL;
};

int io_reference::getIndex(int arraydim, int indexdim)
{
  return index_level[arraydim][indexdim];
};

void io_reference::setStride(int arraydim, int indexdim, SgExpression * expr)
{
  stride[arraydim][indexdim]=expr;
};

SgExpression * io_reference::getStride(int arraydim, int indexdim)
{
  return stride[arraydim][indexdim];
};

void io_reference::setOffset(int arraydim, int indexdim, SgExpression * expr)
{
  offset[arraydim][indexdim]=expr;
};

SgExpression * io_reference::getOffset(int arraydim, int indexdim)
{
  return offset[arraydim][indexdim];
};

int io_reference::hasIndexOnDim(int dim)
{
  for (int i=0; i<MAX_INDEX_DIM; i++) {
    if (index_level[dim][i]!=0)
      return 1;
  }
  return 0;
};

int io_reference::indexSequenceOnDim(int dim)
{
  for (int i=0; i<MAX_INDEX_DIM; i++)
    if (index_level[dim][i]!=0)
      return i;
  // return the first one;
  return (printf("This is a scalar dim.\n"));
}

int io_reference::evaluatable()
{
  for (int i=0; i<MAX_ALIGN_DIM; i++) {
    for (int j=0; j<MAX_INDEX_DIM; j++) {
      if (stride[i][j]!=NULL)
	if(!stride[i][j]->isInteger()) {
	  return(0);
	};

      if (offset[i][j]!=NULL)
	if(!stride[i][j]->isInteger()) {
	  return(0);
	}
	  
    }
  }
  return (1);
}

int io_reference::checkRegular()
{
  switch (expr->variant()) {
  case VAR_REF: 
    return (1);
    break;
  case ARRAY_REF:
    {
      SgArrayRefExp * array = (SgArrayRefExp *) expr;

      for (int i=0; i<array->numberOfSubscripts(); i++) {
	int indexcounter = 0;
	for (int j=0; j<MAX_INDEX_DIM; j++)
	  if (index_level[i][j]!=0)
	    if (index_level[i][j]!=1)
	      return 0;
	    else
	      indexcounter++;
	if (indexcounter>1)
	  return 0;
      }

      return 1;
      break;
    };
  default:
    printf("Not touched forall reference!\n");
  }
  return 0;
};

void io_reference::show()
{

};

//------------------------;

void fill_impdo_header(Impdo_header_def * ihdef, SgExpression * exp)
{
  switch (exp->variant()) {

  case EXPR_LIST: {
    SgExprListExp * exprlist=(SgExprListExp *)exp;
    for (int i=0; i<exprlist->length(); i++) {
      fill_impdo_header(ihdef, exprlist->elem(i));
    };
    break;
  };

  case IMPLIED_DO: {
    SgImpliedDoExp * impdo = (SgImpliedDoExp *) exp; 
    ihdef->setIndexSymbol(impdo->index());

    for (int i=0; i<impdo->numberOfTriplet(); i++) {
      ihdef->setIndexTriplet(i, impdo->triplet(i));
    }

    SgExpression * list = impdo->list();    
    fill_impdo_header(ihdef, list);
    
    break;
  };

  default:
    break;
  }
};

void get_impdo_header(SgInputOutputStmt * io)
{
  SgExpression * item = io->itemList();

  switch (item->variant()) {
  case EXPR_LIST:
    {
      if (item->operand(1)->variant()==IMPLIED_DO) {
	Impdo_header_def * ihdef = new Impdo_header_def();
	fill_impdo_header(ihdef, item->operand(1));
	io->addAttribute(IHDER_Def,(void *) ihdef, sizeof(Impdo_header_def));
      }
      else
	printf("not touched\n");
      break;
    }
  default:
    printf("not touched in io item!\n");
    break;
  }
}

void add_impdo_reference(SgStatement * s, SgExpression *exp)
{
  switch (exp->variant()) {
  case IMPLIED_DO: {
    SgImpliedDoExp * impdo = (SgImpliedDoExp *) exp; 
    add_impdo_reference(s, impdo->list());
    break;
  }
  
  case EXPR_LIST: {
    SgExprListExp * exprlist=(SgExprListExp *)exp;
    for (int i=0; i<exprlist->length(); i++) {
      add_impdo_reference(s, exprlist->elem(i));
    }
    break;
  }

  case ADD_OP:
  case MINUS_OP:
  case MULT_OP:
  case DIV_OP:
  case MOD_OP: {
    add_impdo_reference(s, exp->operand(1));
    add_impdo_reference(s, exp->operand(2));
    break;
  }

  case ARRAY_REF: {
    io_reference * iref = new io_reference();
    iref->setExpression(exp);
    s->addAttribute(Io_Ref,(void *)iref, sizeof(io_reference));
    break;
  }

  case INT_VAL:
  case STRING_VAL:
  case VAR_REF: {
    // do nothing;
    break;
  }
  
  default:
    printf("not touched in io item!\n");
    break;

  }
}
  
void get_impdo_reference(SgInputOutputStmt * print)
{
  add_impdo_reference(print, print->itemList()->operand(1));
};


void index_impdo_reference(SgInputOutputStmt * print)
{
  if (print->numberOfAttributes(IHDER_Def)!=0) {
    Impdo_header_def * ihdef = (Impdo_header_def *) print->getAttribute(0, IHDER_Def)->getAttributeData();

    for (int i=0; i<ihdef->numberOfIndex(); i++) {

      SgSymbol * index = ihdef->indexSymbol(i);

      for (int j=0; j<print->numberOfAttributes(Io_Ref); j++) {
	io_reference * iref = (io_reference *) print->getAttribute(j, Io_Ref)->getAttributeData();

	SgArrayRefExp * arrayexp = (SgArrayRefExp *) iref->expression();

	for (int k=0; k<arrayexp->numberOfSubscripts();k++) {
	  int indexlevel=getreflevel(index, arrayexp->subscript(k));
	  iref->setIndex(k, i, indexlevel);
	  
	  if (indexlevel==1) {
	    SgExpression * stride = getlinerexp(1, index, arrayexp->subscript(k));
	    SgExpression * offset = getlinerexp(0, index, arrayexp->subscript(k));

	    iref->setStride (k, i, stride);
	    iref->setOffset (k, i, offset);
	  }
	}

      }
    }
  }
};

void analyse_print(SgStatement * s)
{
  SgInputOutputStmt * print = (SgInputOutputStmt *)s;

  //printf ("%d, %d\n", print->specList()->id(), print->itemList()->id());

  get_impdo_header(print);
  get_impdo_reference(print);
  index_impdo_reference(print);

}

void add_grp0(SgInputOutputStmt * print)
{
  SgSymbol * grp = print->getSymbolInCurrentScope(mixstring, "grp", NULL, 0);
  SgExpression * grpexp = new SgExpression(VAR_REF, NULL, NULL, grp);

  {
    SgSymbol * fname = print->getSymbolInCurrentScope(mixstring, "new_group_grid", NULL);
    SgFunctionRefExp * funcexpr = new SgFunctionRefExp (*fname);

    SgValueExp * rank = new SgValueExp(0);
    funcexpr->addArg(*rank);
    
    SgValueExp * processor = new SgValueExp(0);
    funcexpr->addArg(* processor);

    SgAssignStmt * assign = new SgAssignStmt(*grpexp, *funcexpr);

    print->insertStmtBefore(*assign, *print->controlParent());

  }

  {
    SgSymbol * fname = print->getSymbolInCurrentScope(mixstring, "delete_group", NULL);

    SgCallStmt * afuncall = new SgCallStmt(*fname);

    SgExpression *grpexp = new SgExpression(VAR_REF, NULL, NULL,grp);

    afuncall->addArg(*grpexp);
    
    print->insertStmtAfter(*afuncall, *print->controlParent());

  }
}

void create_section(SgInputOutputStmt * print, int counter)
{

  Impdo_header_def * ihdef=NULL;

  if (print->numberOfAttributes(IHDER_Def)!=0) {
    ihdef = (Impdo_header_def *) print->getAttribute(0, IHDER_Def)->getAttributeData();
  };

  io_reference * iref= (io_reference *) ((print->getAttribute(counter, Io_Ref))->getAttributeData());

  SgArrayRefExp * array = (SgArrayRefExp *) iref->expression();  
  SgSymbol * arraysymb  = array->symbol();

  {
    SgSymbol * symb;

    symb= print->getSymbolInCurrentScope(mixstring, "new_array_section", NULL);
    SgFunctionRefExp * funref = new SgFunctionRefExp(*symb);
    
    if (ihdef!=NULL) {
      int dimcounter=0;
      for (int i=0; i<array->numberOfSubscripts(); i++)
	if (iref->hasIndexOnDim(i))
	  dimcounter++;
      SgValueExp * dim = new SgValueExp(dimcounter);
      funref->addArg(*dim);
    }
    else {
      SgValueExp * dim = new SgValueExp(0);
      funref->addArg(*dim);
    }

    SgSymbol * dad = print->getSymbolInCurrentScope(mixstring, "dad", arraysymb->identifier());
    SgExpression * dadexp = new SgExpression(VAR_REF, NULL, NULL, dad);

    funref->addArg(*dadexp);

    symb=print->getSymbolInCurrentScope(mixstring, "sdd", NULL, counter);

    SgExpression * dadsecexp = new SgExpression (VAR_REF, NULL, NULL, symb);

    SgAssignStmt * assign = new SgAssignStmt ( *dadsecexp, *funref);

    print->insertStmtBefore(*assign, *print->controlParent());
  }

  for (int i=0; i<array->numberOfSubscripts(); i++) {
    static int dimcounter =0;

    if (iref->hasIndexOnDim(i)) {
      SgSymbol * symb= print->getSymbolInCurrentScope(mixstring, "set_array_triplet", NULL);
      SgCallStmt * funcall= new SgCallStmt(*symb);

      {
	SgSymbol * symb = print->getSymbolInCurrentScope(mixstring, "sdd", NULL, counter);
	SgExpression * dadsecexp = new SgExpression(VAR_REF, NULL, NULL, symb);

	funcall->addArg(*dadsecexp);
      }
      {
	dimcounter++;
	SgValueExp * secdim = new SgValueExp (dimcounter);
	funcall->addArg(*secdim);
      }
 
      // we need to change the header, so all expression should be a copy;
      ihdef = (Impdo_header_def *) print->getAttribute(0, IHDER_Def)->getAttributeData();
     
      {
	for (int j=0; j<2; j++) {
	  if (ihdef->getIndexTriplet(iref->indexSequenceOnDim(i), j)!=NULL) {
	    SgExpression * triplet = ihdef->getIndexTriplet(iref->indexSequenceOnDim(i), j);

	    SgExpression * newtriplet = NULL;
	    
	    newtriplet=newtriplet+(*iref->getOffset(i, iref->indexSequenceOnDim(i)));

	    newtriplet=newtriplet+((*iref->getStride(i, iref->indexSequenceOnDim(i)))*(triplet->copy()));

	    if (newtriplet->isInteger())
	      newtriplet=new SgValueExp(newtriplet->valueInteger());
				    
	    funcall->addArg(*newtriplet);
	  }
	}
	{
	  //	  if (ihdef->getIndexTriplet(iref->indexSequenceOnDim(i), 2)!=NULL) {
	  //      we don't consider the stride in the header at the moment;
	  if(0) {
	    SgExpression *triplet=ihdef->getIndexTriplet(iref->indexSequenceOnDim(i), 2);

	    SgExpression * newtriplet = new SgExpression (EXPR_LIST);

	    (*newtriplet)=(triplet->copy());

	    if (iref->getStride(i, iref->indexSequenceOnDim(i))!=NULL)
	      newtriplet=newtriplet*(*iref->getStride(i, iref->indexSequenceOnDim(i)));

	    if (newtriplet->isInteger())
	      newtriplet=new SgValueExp(newtriplet->valueInteger());

	    funcall->addArg(*newtriplet);
	  }
	  else{
	    SgExpression * newtriplet = new SgValueExp(1);

	    if (iref->getStride(i, iref->indexSequenceOnDim(i))!=NULL)
	      newtriplet=newtriplet*(*iref->getStride(i, iref->indexSequenceOnDim(i)));

	    if (newtriplet->isInteger())
	      newtriplet=new SgValueExp(newtriplet->valueInteger());

	    funcall->addArg(*newtriplet);
	  }
	}
      }

      {
        SgSymbol * dad = print->getSymbolInCurrentScope(mixstring, "dad", arraysymb->identifier());	
	SgExpression * dadexp = new SgExpression(VAR_REF, NULL, NULL, dad);

	funcall->addArg(*dadexp);
      }
      {
	SgValueExp * dim = new SgValueExp (i+1);
	funcall->addArg(*dim);
      }
      print->insertStmtBefore(*funcall, *print->controlParent());
    }
    else {
      SgSymbol * symb= print->getSymbolInCurrentScope(mixstring, "set_array_scalar", NULL);
      SgCallStmt * funcall= new SgCallStmt(*symb);
      {
	SgSymbol * symb = print->getSymbolInCurrentScope(mixstring, "sdd", NULL, counter);
	SgExpression * dadsecexp = new SgExpression(VAR_REF, NULL, NULL, symb);

	funcall->addArg(*dadsecexp);
      }
      {
	SgExpression * scalar = ( (SgArrayRefExp *) iref->expression() ) ->subscript(i);
	funcall->addArg(*scalar);
      }
      {
        SgSymbol * dad = print->getSymbolInCurrentScope(mixstring, "dad", arraysymb->identifier());
	SgExpression * dadexp = new SgExpression(VAR_REF, NULL, NULL, dad);
	
	funcall->addArg(*dadexp);
      };
      {
	SgValueExp * dim = new SgValueExp (i+1);
	funcall->addArg(*dim);
      }
      print->insertStmtBefore(*funcall, *print->controlParent());
    };
  }

  // delete dad;
  {
    SgSymbol * symb=print->getSymbolInCurrentScope(mixstring, "delete_array", NULL);
    SgCallStmt * funcall= new SgCallStmt(*symb);

    {
      SgSymbol * symb=print->getSymbolInCurrentScope(mixstring, "sdd", NULL, counter);
      SgExpression * dadexp = new SgExpression(VAR_REF, NULL, NULL, symb);
      funcall->addArg(*dadexp);
    };

    print->insertStmtAfter(*funcall, *print->controlParent());
  };

};

void create_array(SgInputOutputStmt * print, int counter)
{

  Impdo_header_def * ihdef=NULL;

  if (print->numberOfAttributes(IHDER_Def)!=0) {
    ihdef = (Impdo_header_def *) print->getAttribute(0, IHDER_Def)->getAttributeData();
  };

  io_reference * iref= (io_reference *) ((print->getAttribute(counter, Io_Ref))->getAttributeData());

  SgArrayRefExp * array = (SgArrayRefExp *) iref->expression();  
  SgSymbol * arraysymb  = array->symbol();

  {
    SgSymbol * symb;

    symb= print->getSymbolInCurrentScope(mixstring, "new_array_data", NULL);
    SgFunctionRefExp * funref = new SgFunctionRefExp(*symb);

    SgType * stype= arraysymb->type();

    SgSymbol * stackname=print->getSymbolInCurrentScope(mixstring, stackNameOfType(stype), NULL);
    SgExpression * stackexp = new SgExpression (VAR_REF, NULL, NULL, stackname);
    funref->addArg(*stackexp);
    
    SgSymbol * typeName=print->getSymbolInCurrentScope(mixstring, nameOfType(stype), NULL);
    SgExpression * typeExp = new SgExpression (VAR_REF, NULL, NULL, typeName);
    funref->addArg(*typeExp);
    
    SgSymbol * sizeName = print->getSymbolInCurrentScope(mixstring, sizeNameOfType(stype), NULL);
    SgExpression * sizeExp = new SgExpression (VAR_REF, NULL, NULL, sizeName);
    funref->addArg(*sizeExp);
    
    if (ihdef!=NULL) {
      int dimcounter=0;
      for (int i=0; i<array->numberOfSubscripts(); i++)
	if (iref->hasIndexOnDim(i))
	  dimcounter++;
      SgValueExp * dim = new SgValueExp(dimcounter);
      funref->addArg(*dim);
    }
    else {
      SgValueExp * dim = new SgValueExp(0);
      funref->addArg(*dim);
    }
    
    SgSymbol * grp=print->getSymbolInCurrentScope(mixstring, "grp0", NULL);
    SgExpression * grpexp = new SgExpression (VAR_REF, NULL, NULL, grp);
    funref->addArg(*grpexp);

    symb=print->getSymbolInCurrentScope(mixstring, "tdd", NULL, counter);
    SgExpression * dadsecexp = new SgExpression (VAR_REF, NULL, NULL, symb);
    SgAssignStmt * assign = new SgAssignStmt ( *dadsecexp, *funref);

    print->insertStmtBefore(*assign, *print->controlParent());
  }

  for (int i=0; i<array->numberOfSubscripts(); i++) {

    static dimcounter=0;

    if (iref->hasIndexOnDim(i)) {
      dimcounter++;

      SgSymbol * fname = print->getSymbolInCurrentScope(mixstring, "set_array_collapse", NULL);
      SgCallStmt * func = new SgCallStmt (*fname);
	    
      SgSymbol * dad = print ->getSymbolInCurrentScope(mixstring, "tdd", NULL, counter);
      SgExpression * dadexp = new SgExpression (VAR_REF, NULL, NULL, dad);

      func->addArg(*dadexp);

      SgValueExp * dimention = new SgValueExp(dimcounter);
	    
      func->addArg(*dimention);

      ihdef = (Impdo_header_def *) print->getAttribute(0, IHDER_Def)->getAttributeData();

      int indexcounter=iref->indexSequenceOnDim(i);

      {
	SgValueExp *zero = new SgValueExp(0);
	
	func->addArg(*zero);
      }

      {
	SgExpression *upper = new SgExpression(EXPR_LIST);

	(*upper)=(ihdef->getIndexTriplet(indexcounter, 1)->copy());
	(*upper)=(*upper)-(ihdef->getIndexTriplet(indexcounter, 0)->copy());

	if (upper->isInteger())
	  upper= new SgValueExp(upper->valueInteger());

	func->addArg(*upper);
      }

      print->insertStmtBefore(*func, *print->controlParent());

    }
  }

  // delete dad;
  {
    SgSymbol * symb=print->getSymbolInCurrentScope(mixstring, "delete_array", NULL);
    SgCallStmt * funcall= new SgCallStmt(*symb);

    {
      SgSymbol * symb=print->getSymbolInCurrentScope(mixstring, "tdd", NULL, counter);
      SgExpression * dadexp = new SgExpression(VAR_REF, NULL, NULL, symb);
      
      funcall->addArg(*dadexp);
    };

    print->insertStmtAfter(*funcall, *print->controlParent());
  };
  
  // allocate memory;
  {
    SgSymbol * base = print->getSymbolInCurrentScope(mixstring, "tbs", NULL, counter);
    SgExpression * baseexp=new SgExpression(VAR_REF, NULL, NULL, base);

    SgSymbol * franame = print->getSymbolInCurrentScope(mixstring, "real_alloc", NULL);
    SgFunctionRefExp * rafuncexpr = new SgFunctionRefExp (*franame);

    SgSymbol * fsizename = print->getSymbolInCurrentScope(mixstring, "size", NULL);
    SgFunctionRefExp * sizefuncexpr = new SgFunctionRefExp (*fsizename);

    SgSymbol * dadbase = print->getSymbolInCurrentScope(mixstring, "tdd", NULL, counter);
    SgExpression * dadbaseexp=new SgExpression(VAR_REF, NULL, NULL, dadbase);
    sizefuncexpr->addArg(*dadbaseexp);

    rafuncexpr->addArg(*sizefuncexpr);

    SgStatement * alloc = new SgAssignStmt(*baseexp, *rafuncexpr);

    print->insertStmtBefore(*alloc, *print->controlParent());
  }

  // reset base;
  {
    SgSymbol * symb = print->getSymbolInCurrentScope(mixstring, "reset_array_base", NULL);
    SgCallStmt * funcall = new SgCallStmt (*symb);

    SgSymbol * base = print->getSymbolInCurrentScope(mixstring, "tdd", NULL, counter);
    SgExpression * baseexpr=new SgExpression(VAR_REF, NULL, NULL, base);
    funcall->addArg(*baseexpr);

    SgSymbol * rs= print->getSymbolInCurrentScope(mixstring, "real_stack", NULL);
    SgArrayRefExp * stackarray = new SgArrayRefExp(*rs);

    SgSymbol * basesym = print->getSymbolInCurrentScope(mixstring, "tbs", NULL,counter);
    SgExpression * baseexp= new SgExpression(VAR_REF, NULL, NULL, basesym);

    stackarray->addSubscript(*baseexp);
    funcall->addArg(*stackarray);

    print->insertStmtBefore(*funcall, *print->controlParent());
  }

  // call real free;
  {
    SgSymbol * symb = print->getSymbolInCurrentScope(mixstring, "real_free", NULL);
    SgCallStmt * funcall = new SgCallStmt (*symb);
    
    SgSymbol * base = print->getSymbolInCurrentScope(mixstring, "tbs", NULL, counter);
    SgExpression * baseexpr=new SgExpression(VAR_REF, NULL, NULL, base);
    
    funcall->addArg(*baseexpr);
    
    print->insertStmtAfter(*funcall, *print->controlParent());
  }
  
};

void change_reference(SgInputOutputStmt * print, int counter)
{
  io_reference * iref= (io_reference *) ((print->getAttribute(counter, Io_Ref))->getAttributeData());

  SgArrayRefExp * array = (SgArrayRefExp *) iref->expression();  
  SgSymbol * arraysymb  = array->symbol();

  SgSymbol * newsymb = print->getSymbolInCurrentScope(mixstring, stackNameOfType(arraysymb->type()), NULL);

  SgArrayRefExp * newarray = new SgArrayRefExp(*newsymb);

  SgSymbol * basesym = print->getSymbolInCurrentScope(mixstring, "tbs", NULL,counter);
  SgExpression * baseexp= new SgExpression(VAR_REF, NULL, NULL, basesym);

  if (print->numberOfAttributes(IHDER_Def)!=0) {
    Impdo_header_def * ihdef = (Impdo_header_def *) print->getAttribute(0, IHDER_Def)->getAttributeData();

    // since this is a new array, we can get memstride by simple computation;

    SgExpression * memstride = NULL;
    
    for (int i = 0; i< ihdef->numberOfIndex(); i++) {
  
      SgSymbol * indexsym = ihdef->indexSymbol(i);
      SgExpression * newindex = new SgExpression(VAR_REF, NULL, NULL, indexsym);

      newindex = memstride * (*newindex);

      (*baseexp)=(*baseexp)+(*newindex);

      //update memstride in the following;

      SgExpression * lb=ihdef->getIndexTriplet(i, 0);
      SgExpression * ub=ihdef->getIndexTriplet(i, 1);

      memstride = memstride * ((*ub)-(*lb));

      if (memstride->isInteger())
	memstride = new SgValueExp(memstride->valueInteger());
    }
  }

  newarray->addSubscript(*baseexp);

  array->replace(*newarray);

};

void add_remap(SgInputOutputStmt * print, int counter)
{
  SgSymbol * symb;
  
  symb=print->getSymbolInCurrentScope(mixstring, "remap", NULL);
  
  SgCallStmt * funcall= new SgCallStmt(*symb);

  symb=print->getSymbolInCurrentScope(mixstring, "tdd", NULL, counter);
  SgExpression * dadexp = new SgExpression(VAR_REF, NULL, NULL, symb);
  funcall->addArg(*dadexp);

  symb=print->getSymbolInCurrentScope(mixstring, "sdd", NULL , counter);
  SgExpression * dadsecexp = new SgExpression(VAR_REF, NULL, NULL, symb);
  funcall->addArg(*dadsecexp);

  print->insertStmtBefore(*funcall, *print->controlParent());
  
};

void change_io_reference(SgInputOutputStmt * print, int counter) 
{
  Impdo_header_def * ihdef=NULL;

  if (print->numberOfAttributes(IHDER_Def)!=0) {
    ihdef = (Impdo_header_def *) print->getAttribute(0, IHDER_Def)->getAttributeData();
  };

  io_reference * iref = (io_reference *) print->getAttribute(counter, Io_Ref)->getAttributeData();

  //printf("the io reference is %d , %s\n", iref->expression()->id(),iref->expression()->symbol()->identifier() );

  if ((iref->expression()->symbol()->numberOfAttributes(Align_Def)==0)
      &&(iref->expression()->symbol()->numberOfAttributes(Inherit_Def)==0)) {
    // local no explicit distributed array, do nothing;
  }
  else {
    if (iref->checkRegular()==1) {
      create_section(print, counter);
      create_array(print, counter);
      add_remap(print, counter);
      change_reference(print, counter);
    }
    else {
      printf("not finished yet!\n");
    }
  }
}

void change_impdo_header(SgInputOutputStmt * print)
{
  if (print->numberOfAttributes(IHDER_Def)!=0) {
    Impdo_header_def * ihdef = (Impdo_header_def *) print->getAttribute(0, IHDER_Def)->getAttributeData();

    for (int i = 0; i< ihdef->numberOfIndex(); i++) {

      SgExpression * lb=ihdef->getIndexTriplet(i, 0);
      SgExpression * ub=ihdef->getIndexTriplet(i, 1);
      SgExpression * stride =ihdef->getIndexTriplet(i, 2);

      SgExpression * newlb = new SgValueExp(0);

      //SgExpression * newst = new SgValueExp(1);

      SgExpression * newub = new SgExpression (EXPR_LIST);
      
      // we do not change stride temporarily, in case u-l can not divide s;
      
      if (stride!=NULL)
	(* newub) = ((* ub)-(*lb));
      else 
	(* newub) = ((* ub)-(*lb));

      if (newub->isInteger())
	newub = new SgValueExp(newub->valueInteger());

      lb->replace(*newlb);
      ub->replace(*newub);

    }    

  };

}

void change_print(SgInputOutputStmt * print)
{

  for (int i=0; i<print->numberOfAttributes(Io_Ref); i++) {
    change_io_reference(print, i);
  };

  change_impdo_header(print);

  SgSymbol * on = print->getSymbolInCurrentScope(mixstring, "on", NULL);
  SgFunctionRefExp * onexp = new SgFunctionRefExp(*on);

  SgSymbol * grp = print->getSymbolInCurrentScope(mixstring, "grp", NULL, 0);
  SgExpression * grpexp = new SgExpression(VAR_REF, NULL, NULL, grp);

  onexp->addArg(*grpexp);

  SgIfStmt * ifstmt = new SgIfStmt(*onexp, print->copy());

  print->insertStmtBefore(*ifstmt, *print->controlParent());

};

void trans_print(SgStatement * s)
{
  SgInputOutputStmt * print = (SgInputOutputStmt *)s;

  add_grp0(print);

  change_print(print);
  
}









